#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2015-2020 Óscar García Amor <ogarcia@connectical.com>
#
# Distributed under terms of the GNU GPLv3 license.

from trellowarrior.exceptions import ClientError
from trello import TrelloClient as Client
from trello.exceptions import ResourceUnavailable

import logging

logger = logging.getLogger(__name__)

class TrelloClient:
    def __init__(self, api_key, api_secret, token, token_secret):
        self.trello_client = Client(api_key=api_key, api_secret=api_secret, token=token, token_secret=token_secret)
        self._uid = None
        self._project = None
        self._board = None
        self._lists = None
        self._board_labels = None
        self._lists_filter = None
        self._only_my_cards = False

    @property
    def whoami(self):
        """
        Get my Trello UID

        :return: my Trello UID
        :rtype: string
        """
        if self._uid is None:
            self._uid = self.trello_client.get_member('me').id
        return self._uid

    def project(self, project):
        """
        Set the class working project

        :param project: TelloWarrior project object
        """
        if self._project == None or self._project.name != project.name:
            self._board = self.get_board(project.trello_board_name)
            self._lists = self.get_lists()
            self._board_labels = self.get_board_labels()
            self._lists_filter = project.trello_lists_filter
            self._only_my_cards = project.only_my_cards

    def get_board(self, board_name):
        """
        Get a open Trello board from name, if it does not exist create it

        :param board_name: the board name
        :return: a Tello board
        :rtype: Trello board object
        """
        for trello_board in self.trello_client.list_boards(board_filter='open'):
            if trello_board.name == board_name and not trello_board.closed:
                logger.debug('Trello board {} found'.format(board_name))
                return trello_board
        logger.debug('Creating Trello board {}'.format(board_name))
        return self.trello_client.add_board(board_name)

    def get_lists(self):
        """
        Get the open lists of a Trello board

        :return: a list of Trello list objects
        :rtype: list
        """
        return self._board.open_lists()

    def get_list(self, list_name):
        """
        Get a Trello list from list name, if it does not exist create it

        :param list_name: the list name
        :return: a Tello list
        :rtype: Trello list object
        """
        if self._lists == None:
            raise ClientError('get_list')
        for trello_list in self._lists:
            if trello_list.name == list_name:
                logger.debug('Trello list {} found'.format(list_name))
                return trello_list
        logger.debug('Creating Trello list {}'.format(list_name))
        trello_list = self._board.add_list(list_name)
        self._lists.append(trello_list) # Update _lists with new list
        return trello_list

    def get_board_labels(self):
        """
        Get the labels of a Trello board

        :param board_name: the board name
        :return: a list of Trello label objects
        :rtype: list
        """
        return self._board.get_labels()

    def get_board_label(self, label_name):
        """
        Get a Trello board label from label name, if it does not exist create it

        :param label_name: the label name
        :return: a Tello label
        :rtype: Trello label object
        """
        if self._board_labels == None:
            raise ClientError('get_board_label')
        for board_label in self._board_labels:
            if board_label.name == label_name:
                logger.debug('Trello board label {} found'.format(label_name))
                return board_label
        logger.debug('Creating Trello board label {}'.format(label_name))
        board_label = self._board.add_label(label_name, 'black')
        self._board_labels.append(board_label) # Update _board_labels with new label
        return board_label

    def get_cards_dict(self):
        """
        Get all cards of a list of Trello lists in a dictionary

        :return: a dict with Cards
        :rtype: dict
        """
        trello_cards_dict = {}
        if self._lists_filter is not None:
            trello_lists = filter(lambda trello_list: trello_list.name not in self._lists_filter, self._lists)
        for trello_list in trello_lists:
            logger.debug('Getting Trello cards of list {}'.format(trello_list.name))
            trello_cards_dict[trello_list.name] = trello_list.list_cards()
            if self._only_my_cards:
                trello_cards_dict[trello_list.name] = filter(lambda trello_card: self.whoami in trello_card.member_ids, trello_cards_dict[trello_list.name])
        return trello_cards_dict

    def delete_trello_card(self, trello_card_id):
        """
        Delete (forever) a Trello card by ID

        :param trello_card_id: ID of Trello card
        """
        try:
            self.trello_client.get_card(trello_card_id).delete()
        except ResourceUnavailable:
            logger.warning('Cannot find Trello card with ID {} deleted in Task Warrior. Maybe you also deleted it in Trello?'.format(trello_card_id))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2015-2020 Óscar García Amor <ogarcia@connectical.com>
#
# Distributed under terms of the GNU GPLv3 license.

class InvalidOperation(Exception):
    def __init__(self, operation):
        self.operation = operation
        self.message = 'must open config file first'

    def __str__(self):
        return 'Cannot {}, {}'.format(self.operation, self.message)

class ClientError(InvalidOperation):
    def __init__(self, operation):
        self.operation = operation
        self.message = 'must set project first'



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2015-2020 Óscar García Amor <ogarcia@connectical.com>
#
# Distributed under terms of the GNU GPLv3 license.

from trellowarrior.exceptions import InvalidOperation
from configparser import RawConfigParser, NoOptionError

import logging
import os

logger = logging.getLogger(__name__)

class ConfigEditor:
    def __init__(self):
        self.config_parser = RawConfigParser()
        self.config_file = None

    def open(self, config_file):
        """
        Open and read a config file

        :param config_file: config file path
        """
        logger.debug('Opening {}'.format(config_file))
        self.config_file = config_file
        if os.access(config_file, os.R_OK):
            logger.debug('Parsing config file')
            self.config_parser.read(config_file)

    def read(self, section, option, fallback=None):
        """
        Returns a config option value from config file

        :param section: section where the option is stored
        :param option: option name
        :param fallback: (optional) fallback value
        :return: a config option value
        :rtype: string
        """
        if self.config_file == None:
            raise InvalidOperation('read')
        if fallback is None:
            return self.config_parser.get(section, option)
        else:
            return self.config_parser.get(section, option, fallback=fallback)

    def readboolean(self, section, option, fallback=False):
        """
        Returns a boolean config option value from config file

        :param section: section where the option is stored
        :param option: option name
        :param fallback: (optional) fallback value
        :return: a config option value
        :rtype: boolean
        """
        if self.config_file == None:
            raise InvalidOperation('readboolean')
        return self.config_parser.getboolean(section, option, fallback=fallback)

    def write(self, section, option, value):
        """
        Write a config option value in config object

        :param section: section where the option is stored
        :param option: option name
        :param value: option value
        """
        if self.config_file == None:
            raise InvalidOperation('write')
        if section != 'DEFAULT' and not self.config_parser.has_section(section):
            logger.debug('Adding new section {}'.format(section))
            self.config_parser.add_section(section)
        logger.debug('Adding {}.{} with value {}'.format(section, option, value))
        self.config_parser.set(section, option, value)

    def remove(self, section, option):
        """
        Remove a config option in config object

        :param section: section where the option is stored
        :param option: option name
        :return: True if option is removed, False if not exist
        :rtype: boolean
        """
        if self.config_file == None:
            raise InvalidOperation('remove')
        logger.debug('Removing {}.{}'.format(section, option))
        option_removed = self.config_parser.remove_option(section, option)
        if section != 'DEFAULT' and option_removed:
            if self.config_parser.items(section) == self.config_parser.items('DEFAULT'):
                logger.debug('Removing empty section {}'.format(section))
                self.config_parser.remove_section(section)
        return option_removed

    def remove_project(self, project):
        """
        Remove a project (config section in config object)

        :param project: section name
        :return: True if section is removed, False if not exist
        :rtype: boolean
        """
        if self.config_file == None:
            raise InvalidOperation('remove')
        logger.debug('Removing {}'.format(project))
        return self.config_parser.remove_section(project)

    def list(self):
        """
        List config sections

        :return: list of projects (sections in config)
        :rtype: list
        """
        if self.config_file == None:
            raise InvalidOperation('list')
        return self.config_parser.sections()

    def list_enabled_projects(self):
        """
        Get the list of enabled projects

        :return: list of enabled projects
        :rtype: list
        """
        if self.config_file == None:
            raise InvalidOperation('list_enabled_projects')
        try:
            return self.config_parser.get('DEFAULT', 'sync_projects').split()
        except NoOptionError:
            return []

    def enable_project(self, project):
        """
        Enable a project adding it to sync_projects

        :param project: project name
        """
        if self.config_file == None:
            raise InvalidOperation('enable_project')
        logger.debug('Enabling project {}'.format(project))
        enabled_projects = self.list_enabled_projects()
        enabled_projects.append(project)
        enabled_projects.sort()
        self.config_parser.set('DEFAULT', 'sync_projects', ' '.join(enabled_projects))

    def disable_project(self, project):
        """
        Disable a project removing it from sync_projects

        :param project: project name
        :return: True if project is disabled, False if not
        :rtype: boolean
        """
        if self.config_file == None:
            raise InvalidOperation('disable_project')
        logger.debug('Disabling project {}'.format(project))
        enabled_projects = self.list_enabled_projects()
        try:
            enabled_projects.remove(project)
            self.config_parser.set('DEFAULT', 'sync_projects', ' '.join(enabled_projects))
            return True
        except ValueError:
            logger.debug('Nothing to do, {} is not in enabled projects'.format(project))
            return False

    def has_project(self, project):
        """
        Check if a project (a section in config) is present

        :param project: section name
        :return: True if section exists, False if not
        :rtype: boolean
        """
        if self.config_file == None:
            raise InvalidOperation('has_project')
        return self.config_parser.has_section(project)

    def has_project_enabled(self, project):
        """
        Check if a project is enabled

        :param project: project name
        :return: True if project is enabled, False if not
        :rtype: boolean
        """
        if self.config_file == None:
            raise InvalidOperation('has_project_enabled')
        return True if project in self.list_enabled_projects() else False

    def save(self):
        """
        Save the config object in config file
        """
        if self.config_file == None:
            raise InvalidOperation('save')
        logger.debug('Saving config in config file')
        with open(self.config_file, 'w') as configfile:
            self.config_parser.write(configfile)
        self.config_file = None

    def clean(self):
        """
        Cleans the config editor
        """
        logger.debug('Cleaning config editor')
        self.config_parser = RawConfigParser()
        self.config_file = None

config_editor = ConfigEditor()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2015-2020 Óscar García Amor <ogarcia@connectical.com>
#
# Distributed under terms of the GNU GPLv3 license.

class TrelloWarriorProject:
    def __init__(self, name, taskwarrior_project_name, trello_board_name, **kwargs):
        self.name = name
        self.taskwarrior_project_name = taskwarrior_project_name
        self.trello_board_name = trello_board_name
        self.trello_todo_list = kwargs.get('trello_todo_list', 'To Do')
        self.trello_doing_list = kwargs.get('trello_doing_list', 'Doing')
        self.trello_done_list = kwargs.get('trello_done_list', 'Done')
        self.trello_lists_filter = kwargs.get('trello_lists_filter', None)
        self.only_my_cards = kwargs.get('only_my_cards', False)

    def __repr__(self):
        return '<TrelloWarriorProject {}>'.format(self.name)

    def __str__(self):
        return str(self.__dict__)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2015-2020 Óscar García Amor <ogarcia@connectical.com>
#
# Distributed under terms of the GNU GPLv3 license.

from trellowarrior.exceptions import ClientError
from tasklib.backends import TaskWarrior as Client
from tasklib.task import Task

class TaskwarriorClient:
    def __init__(self, taskrc_location, data_location):
        self.taskwarrior_client = Client(taskrc_location=taskrc_location, data_location=data_location)
        self._project = None

    def project(self, project):
        """
        Set the class working project

        :param project: TelloWarrior project object
        """
        self._project = project.taskwarrior_project_name

    def new_task(self):
        """
        Create a new empty task

        :return: a empty task
        :rtype: Taskwarrior task object
        """
        return Task(self.taskwarrior_client)

    def get_tasks_ids_set(self):
        """
        Get a IDs set of pending and completed tasks

        :return: a set of IDs
        :rtype: set
        """
        if self._project == None:
            raise ClientError('get_tasks_ids_set')
        pending_tasks_ids = set((task['trelloid'] for task in self.taskwarrior_client.tasks.pending().filter(project=self._project)))
        completed_tasks_ids = set((task['trelloid'] for task in self.taskwarrior_client.tasks.completed().filter(project=self._project)))
        return pending_tasks_ids | completed_tasks_ids

    def get_pending_tasks(self, trelloid=None):
        """
        Get a list of pending tasks in a Taskwarrior project

        :param trelloid: trelloid filter (None by default)
        :return: a list of Taskwarrior tasks objects
        :rtype: list
        """
        if self._project == None:
            raise ClientError('get_pending_tasks')
        return self.taskwarrior_client.tasks.pending().filter(project=self._project, trelloid=trelloid)

    def get_completed_tasks(self, trelloid=None):
        """
        Get a list of completed tasks in a Taskwarrior project

        :param trelloid: trelloid filter (None by default)
        :return: a list of Taskwarrior tasks objects
        :rtype: list
        """
        if self._project == None:
            raise ClientError('get_completed_tasks')
        return self.taskwarrior_client.tasks.completed().filter(project=self._project, trelloid=trelloid)

    def get_deleted_tasks(self):
        """
        Get a list of deleted tasks in a Taskwarrior project

        :return: a list of Taskwarrior tasks objects
        :rtype: list
        """
        if self._project == None:
            raise ClientError('get_completed_tasks')
        return self.taskwarrior_client.tasks.filter(project=self._project, status='deleted')

    def get_task_by_trello_id(self, trello_id):
        """
        Get a task by Trello ID
        Trello ID must be unique, if not this raise an error

        :param trello_id: Trello card ID
        :return: a Taskwarrior task or None if task not Found
        :rtype: Taskwarrior task object
        """
        tasks = self.taskwarrior_client.tasks.filter(trelloid=trello_id)
        tasks_count = len(tasks)
        if tasks_count == 0:
            return None
        elif tasks_count == 1:
            return tasks[0]
        else:
            raise ValueError('Duplicated Trello ID {} in Taskwarrior tasks. Trello IDs must be unique, please fix it before sync'.format(trello_id))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2015-2020 Óscar García Amor <ogarcia@connectical.com>
#
# Distributed under terms of the GNU GPLv3 license.

from trellowarrior.trellowarriorproject import TrelloWarriorProject
from configparser import RawConfigParser, NoOptionError

import logging
import os

logger = logging.getLogger(__name__)

class Config:
    """
    Configure TrelloWarrior from various sources (with increasing priority)
        - Defaults
        - Config file
        - Supplied config
    """

    def __init__(self):
        # Configure defaults
        self.config_file = None
        self.taskwarrior_taskrc_location = None
        self.taskwarrior_data_location = None
        self.trello_api_key = None
        self.trello_api_secret = None
        self.trello_token = None
        self.trello_token_secret = None
        self.sync_projects = []

    def configure(self, **kwargs):
        # Get config file location
        self.config_file = kwargs.get('config_file', None)
        # No config file passed, search for it in default locations
        if self.config_file == None:
            user_home = os.path.expanduser('~')
            config_home = os.environ.get('XDG_CONFIG_HOME', os.path.join(user_home, '.config'))
            defaults_config_files = [ './trellowarrior.conf',
                    os.path.join(user_home, '.trellowarrior.conf'),
                    os.path.join(config_home, 'trellowarrior/trellowarrior.conf') ]
            config_files = [file for file in defaults_config_files if os.access(file, os.R_OK)]
            self.config_file = config_files[0] if config_files != [] else None

        if self.config_file == None:
            # No config file exists, use XDG_CONFIG location
            self.config_file = os.path.join(config_home, 'trellowarrior/trellowarrior.conf')
            logger.debug('Using {} as config file'.format(self.config_file))
        elif kwargs.get('parse_config_file', True):
            logger.debug('Using {} as config file'.format(self.config_file))
            # Parse the config
            config_parser = RawConfigParser()
            config_parser.read(self.config_file)

            # Get the TaskWarrior info from config
            self.taskwarrior_taskrc_location = config_parser.get('DEFAULT', 'taskwarrior_taskrc_location', fallback='~/.taskrc')
            self.taskwarrior_data_location = config_parser.get('DEFAULT', 'taskwarrior_data_location', fallback='~/.task')

            # Get the auth info from config
            MandatoryExit = lambda option: SystemExit('Missing mandatory entry \'{}\' in config file'.format(option))
            try:
                self.trello_api_key = config_parser.get('DEFAULT', 'trello_api_key')
            except NoOptionError:
                raise MandatoryExit('trello_api_key')
            try:
                self.trello_api_secret = config_parser.get('DEFAULT', 'trello_api_secret', fallback=None)
            except NoOptionError:
                raise MandatoryExit('trello_api_secret')
            try:
                self.trello_token = config_parser.get('DEFAULT', 'trello_token', fallback=None)
            except NoOptionError:
                raise MandatoryExit('trello_token')
            try:
                self.trello_token_secret = config_parser.get('DEFAULT', 'trello_token_secret', fallback=None)
            except NoOptionError:
                raise MandatoryExit('trello_token_secret')

            # Get the projects to sync
            projects = kwargs.get('projects') if kwargs.get('projects', []) != [] else config_parser.get('DEFAULT', 'sync_projects', fallback='').split()
            for project in projects:
                if config_parser.has_section(project):
                    if not config_parser.has_option(project, 'taskwarrior_project_name') and not config_parser.has_option(project, 'tw_project_name'):
                        logger.warning('Missing mandatory config option \'taskwarrior_project_name\' for sync project \'{}\', ignoring it'.format(project))
                    elif not config_parser.has_option(project, 'trello_board_name'):
                        logger.warning('Missing mandatory config option \'trello_board_name\' for sync project \'{}\', ignoring it'.format(project))
                    else:
                        logger.debug('Configuring project: {}'.format(project))
                        taskwarrior_project_name = config_parser.get(project, 'taskwarrior_project_name', fallback=None)
                        if taskwarrior_project_name is None:
                            logger.warning('Deprecated option \'tw_project_name\' found in \'{}\' project config, you must change it to \'taskwarrior_project_name\' to avoid problems in next release'.format(project))
                            taskwarrior_project_name = config_parser.get(project, 'tw_project_name')
                        lists_filter = config_parser.get(project, 'trello_lists_filter', fallback='').split(',')
                        todo_list = config_parser.get(project, 'trello_todo_list', fallback='To Do')
                        doing_list = config_parser.get(project, 'trello_doing_list', fallback='Doing')
                        done_list = config_parser.get(project, 'trello_done_list', fallback='Done')
                        FilterWarning = lambda basic_list: logger.warning('Cannot add {} list to lists filter, removing it'.format(basic_list))
                        if todo_list in lists_filter:
                            FilterWarning('todo')
                            lists_filter.remove(todo_list)
                        if doing_list in lists_filter:
                            FilterWarning('doing')
                            lists_filter.remove(doing_list)
                        if done_list in lists_filter:
                            FilterWarning('done')
                            lists_filter.remove(done_list)
                        try:
                            only_my_cards = config_parser.getboolean(project, 'only_my_cards', fallback=False)
                        except ValueError:
                            only_my_cards = False
                            logger.warning('Option \'only_my_cards\' is misconfigured in project \'{}\', ignoring it'.format(project))
                        self.sync_projects.append(TrelloWarriorProject(project,
                            taskwarrior_project_name,
                            config_parser.get(project, 'trello_board_name'),
                            trello_todo_list = todo_list,
                            trello_doing_list = doing_list,
                            trello_done_list = done_list,
                            trello_lists_filter = lists_filter,
                            only_my_cards = only_my_cards))
                else:
                    logger.warning('Missing config section for sync project \'{}\', ignoring it'.format(sync_project))
            if self.sync_projects == []:
                raise SystemExit('No projects to sync')

        # Settings from kwargs
        for key, value in kwargs.items():
            if hasattr(self, key) and key != 'config_file':
                setattr(self, key, value)

    def __str__(self):
        return str(self.__dict__)

config = Config()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2015-2020 Óscar García Amor <ogarcia@connectical.com>
#
# Distributed under terms of the GNU GPLv3 license.

from trellowarrior.config import config
from trellowarrior.configeditor import config_editor
from configparser import NoOptionError

import logging
import sys

logger = logging.getLogger(__name__)

def config_project_list(args):
    # Open config
    config_editor.open(config.config_file)

    # Every config section is a project
    projects = config_editor.list()

    if projects != []:
        projects.sort()
        enabled_projects = config_editor.list_enabled_projects()
        for project in projects:
            # List projects
            status = 'Enabled' if project in enabled_projects else 'Disabled'
            sys.stdout.write('Project \'{}\': {}\n'.format(project, status))
    else:
        sys.stdout.write('No projects declared in config file\n')

def config_project_add(args):
    # Open config
    config_editor.open(config.config_file)

    if config_editor.has_project(args.name):
        sys.stderr.write('A project with name \'{}\' already exists, please use modify command to edit it\n'.format(args.name))
        sys.exit(1)

    if ' ' in args.name:
        sys.stderr.write('A project name must not contain spaces\n')
        sys.exit(1)

    # Add new project
    config_editor.write(args.name, 'taskwarrior_project_name', args.taskwarrior)
    config_editor.write(args.name, 'trello_board_name', args.trello)
    config_editor.write(args.name, 'trello_todo_list', args.todo)
    config_editor.write(args.name, 'trello_doing_list', args.doing)
    config_editor.write(args.name, 'trello_done_list', args.done)
    if args.filter is not None:
        config_editor.write(args.name, 'trello_lists_filter', args.filter)
    if args.only_my_cards:
        config_editor.write(args.name, 'only_my_cards', True)

    if not args.disabled:
        # Add project to enabled projects
        config_editor.enable_project(args.name)

    config_editor.save()
    logger.info('Added {} project \'{}\''.format('disabled' if args.disabled else 'enabled', args.name))

def config_project_modify(args):
    # Check if user provides any option
    if not any([args.taskwarrior, args.trello, args.todo, args.doing, args.done, args.filter, args.only_my_cards]):
        sys.stderr.write('Must provide an option to modify\n')
        sys.exit(1)

    # Open config
    config_editor.open(config.config_file)

    if not config_editor.has_project(args.name):
        sys.stderr.write('Project \'{}\' does not exist in config file\n'.format(args.name))
        sys.exit(1)

    # Modify options
    if args.taskwarrior is not None:
        config_editor.write(args.name, 'taskwarrior_project_name', args.taskwarrior)
    if args.trello is not None:
        config_editor.write(args.name, 'trello_board_name', args.trello)
    if args.todo is not None:
        config_editor.write(args.name, 'trello_todo_list', args.todo)
    if args.doing is not None:
        config_editor.write(args.name, 'trello_doing_list', args.doing)
    if args.done is not None:
        config_editor.write(args.name, 'trello_done_list', args.done)
    if args.filter is not None:
        config_editor.write(args.name, 'trello_lists_filter', args.filter)
    if args.only_my_cards is not None:
        config_editor.write(args.name, 'only_my_cards', True if args.only_my_cards == 'yes' else False)

    config_editor.save()
    logger.info('Project \'{}\' modified'.format(args.name))

def config_project_show(args):
    # Open config
    config_editor.open(config.config_file)

    if not config_editor.has_project(args.name):
        sys.stderr.write('Project \'{}\' does not exist in config file\n'.format(args.name))
        sys.exit(1)

    # List options in project
    try:
        sys.stdout.write('Taskwarrior project name: {}\n'.format(config_editor.read(args.name, 'taskwarrior_project_name')))
    except NoOptionError:
        try:
            sys.stdout.write('Taskwarrior project name (Using DEPRECATED option tw_project_name): {}\n'.format(config_editor.read(args.name, 'tw_project_name')))
        except NoOptionError:
            sys.stdout.write('Warning: missing Taskwarrior project name\n')
    try:
        sys.stdout.write('Trello board name: {}\n'.format(config_editor.read(args.name, 'trello_board_name')))
    except NoOptionError:
        sys.stdout.write('Warning: missing Trello board name\n')
    sys.stdout.write('Trello To Do list: {}\n'.format(config_editor.read(args.name, 'trello_todo_list', 'To Do')))
    sys.stdout.write('Trello Doing list: {}\n'.format(config_editor.read(args.name, 'trello_doing_list', 'Doing')))
    sys.stdout.write('Trello Done list: {}\n'.format(config_editor.read(args.name, 'trello_done_list', 'Done')))
    trello_lists_filter = config_editor.read(args.name, 'trello_lists_filter', '')
    if trello_lists_filter != '':
        sys.stdout.write('Trello lists filter: {}\n'.format(trello_lists_filter))
    try:
        sys.stdout.write('Sync only my cards: {}\n'.format(config_editor.readboolean(args.name, 'only_my_cards')))
    except ValueError:
        sys.stdout.write('Warning: misconfigured only_my_cards option\n')

def config_project_enable(args):
    # Open config
    config_editor.open(config.config_file)

    if not config_editor.has_project(args.name):
        sys.stderr.write('Project \'{}\' does not exist in config file\n'.format(args.name))
        sys.exit(1)

    if config_editor.has_project_enabled(args.name):
        sys.stderr.write('Project \'{}\' is already enabled\n'.format(args.name))
        sys.exit(1)

    # Enable project
    config_editor.enable_project(args.name)
    config_editor.save()
    logger.info('Project \'{}\' enabled'.format(args.name))

def config_project_disable(args):
    # Open config
    config_editor.open(config.config_file)

    # Disable project
    if config_editor.disable_project(args.name):
        config_editor.save()
        logger.info('Project \'{}\' disabled'.format(args.name))
    else:
        sys.stderr.write('Project \'{}\' is already disabled\n'.format(args.name))
        sys.exit(1)

def config_project_remove(args):
    # Open config
    config_editor.open(config.config_file)

    # Remove project
    if config_editor.remove_project(args.name):
        # Disable project too
        config_editor.disable_project(args.name)
        config_editor.save()
        logger.info('Project \'{}\' removed'.format(args.name))
    else:
        sys.stderr.write('Project \'{}\' does not exist in config file\n'.format(args.name))
        sys.exit(1)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2015-2020 Óscar García Amor <ogarcia@connectical.com>
#
# Distributed under terms of the GNU GPLv3 license.

from trellowarrior.clients.trellowarrior import TrelloWarriorClient
from trellowarrior.config import config

def sync(args):
    trellowarrior_client = TrelloWarriorClient(config)
    for project in config.sync_projects:
        trellowarrior_client.sync_project(project)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2015-2020 Óscar García Amor <ogarcia@connectical.com>
#
# Distributed under terms of the GNU GPLv3 license.

from trello.util import create_oauth_token
from trellowarrior.config import config
from configparser import RawConfigParser

import os

def auth(args):
    key = args.api_key or os.environ['TRELLO_API_KEY']
    secret = args.api_key_secret or os.environ['TRELLO_API_SECRET']
    mode = 'r+' if os.path.isfile(config.config_file) else 'w'
    with open(config.config_file, mode) as configfile:
        config_parser = RawConfigParser()
        if mode == 'r+':
            # If file exists read before auth to check it
            config_parser.read_file(configfile)
            configfile.seek(0)

        # Perform auth
        oauth_token = create_oauth_token(expiration=args.expiration, key=key, secret=secret, name=args.name)

        # Write the config
        config_parser.set('DEFAULT', 'trello_api_key', key)
        config_parser.set('DEFAULT', 'trello_api_secret', secret)
        config_parser.set('DEFAULT', 'trello_token', oauth_token['oauth_token'])
        config_parser.set('DEFAULT', 'trello_token_secret', oauth_token['oauth_token_secret'])
        config_parser.write(configfile)



# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2015-2020 Óscar García Amor <ogarcia@connectical.com>
#
# Distributed under terms of the GNU GPLv3 license.

NAME = 'trellowarrior'
VERSION = '1.2.0'
AUTHOR_NAME = 'Óscar García Amor'
AUTHOR_EMAIL = 'ogarcia@connectical.com'
DESCRIPTION = 'Tool to sync Taskwarrior projects with Trello boards'
KEYWORDS = 'board easy project sync task taskwarrior tool trello'
URL = 'https://github.com/ogarcia/trellowarrior'
LICENSE = 'GPLv3'



# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2015-2020 Óscar García Amor <ogarcia@connectical.com>
#
# Distributed under terms of the GNU GPLv3 license.

from trellowarrior import NAME, VERSION

import sys

def version(args):
    sys.stdout.write('{} {}\n'.format(NAME, VERSION))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2015-2020 Óscar García Amor <ogarcia@connectical.com>
#
# Distributed under terms of the GNU GPLv3 license.

from trellowarrior.config import config
from trellowarrior.configeditor import config_editor
from configparser import NoSectionError, NoOptionError

import logging
import sys

logger = logging.getLogger(__name__)

def parse_option(option):
    # Check if is a default option or a section.option
    if '.' in option:
        try:
            section, option = option.split('.')
        except ValueError:
            raise SystemExit('You can only specify an \'option\' or \'section.option\'{}')
    else:
        section, option = 'DEFAULT', option
    logger.debug('Parsed option to section \'{}\' and option \'{}\''.format(section, option))
    return section, option

def config_get(args):
    section, option = parse_option(args.option)

    # Open config
    config_editor.open(config.config_file)

    # Print value
    try:
        sys.stdout.write('{}\n'.format(config_editor.read(section, option)))
    except NoSectionError:
        raise SystemExit('Section \'{}\' not found in config file'.format(section))
    except NoOptionError:
        raise SystemExit('Option \'{}\' not found in section \'{}\''.format(option, section))

def config_set(args):
    section, option = parse_option(args.option)

    # Open config
    config_editor.open(config.config_file)

    # Save value
    config_editor.write(section, option, args.value)
    config_editor.save()
    logger.info('Added \'{}\' with value \'{}\''.format(args.option, args.value))

def config_remove(args):
    section, option = parse_option(args.option)

    # Open config
    config_editor.open(config.config_file)

    # Remove
    try:
        if config_editor.remove(section, option):
            config_editor.save()
            logger.info('Option \'{}\' removed'.format(args.option))
        else:
            sys.stderr.write('Cannot delete \'{}\', does not exist in config file\n'.format(args.option))
            sys.exit(1)
    except NoSectionError:
        raise SystemExit('Section \'{}\' not found in config file'.format(section))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016-2020 Óscar García Amor <ogarcia@connectical.com>
#
# Distributed under terms of the GNU GPLv3 license.

import trellowarrior as project

import os
from setuptools import find_packages
from setuptools import setup

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = project.NAME,
    version = project.VERSION,
    author = project.AUTHOR_NAME,
    author_email = project.AUTHOR_EMAIL,
    description = project.DESCRIPTION,
    license = project.LICENSE,
    keywords = project.KEYWORDS,
    url = project.URL,
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=read('requirements.txt'),
    entry_points={
        'console_scripts': [
            'trellowarrior = trellowarrior.main:main'
        ]
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Utilities',
    ],
    project_urls={
        'Bug Reports': 'https://github.com/ogarcia/trellowarrior/issues',
        'Source': 'https://github.com/ogarcia/trellowarrior',
    },
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2015-2020 Óscar García Amor <ogarcia@connectical.com>
#
# Distributed under terms of the GNU GPLv3 license.

from trellowarrior.clients.taskwarrior import TaskwarriorClient
from trellowarrior.clients.trello import TrelloClient
from trellowarrior.config import config

import logging

logger = logging.getLogger(__name__)

class TrelloWarriorClient:
    def __init__(self, config):
        self.taskwarrior_client = TaskwarriorClient(config.taskwarrior_taskrc_location, config.taskwarrior_data_location)
        self.trello_client = TrelloClient(config.trello_api_key, config.trello_api_secret, config.trello_token, config.trello_token_secret)

    def upload_taskwarrior_task(self, project, taskwarrior_task, trello_list):
        """
        Upload all contents of Taskwarrior task to a Trello list creating a new card and storing cardid

        :param project: TrelloWarrior project object
        :param taskwarrior_task: Taskwarrior task object
        :param trello_list: Trello list object
        """
        new_trello_card = trello_list.add_card(taskwarrior_task['description'])
        if taskwarrior_task['due']:
            new_trello_card.set_due(taskwarrior_task['due'])
        for tag in taskwarrior_task['tags']:
            trello_label = self.trello_client.get_board_label(tag)
            new_trello_card.add_label(trello_label)
        if project.only_my_cards:
            new_trello_card.assign(self.trello_client.whoami)
        taskwarrior_task['trelloid'] = new_trello_card.id
        taskwarrior_task.save()

    def fetch_trello_card(self, project, list_name, trello_card):
        """
        Fetch contents of a Trello card to a new Taskwarrior task

        :param project: TrelloWarrior project object
        :list_name: name of the Trello list where the card is stored
        :param trello_card: Trello card object
        """
        new_taskwarrior_task = self.taskwarrior_client.new_task()
        new_taskwarrior_task['project'] = project.taskwarrior_project_name
        new_taskwarrior_task['description'] = trello_card.name
        if trello_card.due_date:
            new_taskwarrior_task['due'] = trello_card.due_date
        if trello_card.labels:
            for label in trello_card.labels:
                new_taskwarrior_task['tags'].add(label.name)
        new_taskwarrior_task['trelloid'] = trello_card.id
        new_taskwarrior_task['trellolistname'] = list_name
        new_taskwarrior_task.save()
        logger.info('Trello card with ID {} saved as new task in Taskwarrior with ID {}'.format(trello_card.id, new_taskwarrior_task['id']))
        if list_name == project.trello_doing_list:
            new_taskwarrior_task.start()
            logger.info('New task {} kicked to doing list'.format(new_taskwarrior_task['id']))
        if list_name == project.trello_done_list:
            new_taskwarrior_task.done()
            logger.info('New task {} kicked to done list'.format(new_taskwarrior_task['id']))

    def sync_task_card(self, project, list_name, trello_card, taskwarrior_task):
        """
        Sync an existing Taskwarrior task with an existing Trello card

        :param project: TrelloWarrior project object
        :param list_name: name of the Trello list where the card is stored
        :param trello_card: Trello card object
        :param taskwarrior_task: Taskwarrior task object
        """
        taskwarrior_task_modified = False # Change to true to save modification
        # Task description <> Trello card name
        if taskwarrior_task['description'] != trello_card.name:
            if taskwarrior_task['modified'] > trello_card.date_last_activity:
                # Taskwarrior data is newer
                trello_card.set_name(taskwarrior_task['description'])
            else:
                # Trello data is newer
                taskwarrior_task['description'] = trello_card.name
                taskwarrior_task_modified = True
            logger.info('Name of task {} synchronized'.format(taskwarrior_task['id']))
        # Task due <> Trello due
        if taskwarrior_task['due']:
            if not trello_card.due_date or taskwarrior_task['modified'] > trello_card.date_last_activity:
                # No due data in Trello or Taskwarrior data is newer
                trello_card.set_due(taskwarrior_task['due'])
            else:
                # Trello data is newer
                taskwarrior_task['due'] = trello_card.due_date
                taskwarrior_task_modified = True
            logger.info('Due date of task {} synchronized'.format(taskwarrior_task['id']))
        elif trello_card.due_date:
            # No due data in Taskwarrior
            taskwarrior_task['due'] = trello_card.due_date
            taskwarrior_task_modified = True
            logger.info('Due date of task {} synchronized'.format(taskwarrior_task['id']))
        # Task tags <> Trello labels
        trello_card_labels_set = set(trello_card.labels) if trello_card.labels else set()
        trello_card_labels_name_set = set([label.name for label in trello_card_labels_set])
        if taskwarrior_task['tags'] != trello_card_labels_name_set:
            if taskwarrior_task['modified'] > trello_card.date_last_activity:
                # Taskwarrior data is newer
                for tag in taskwarrior_task['tags']:
                    # Get or create label in board
                    trello_label = self.trello_client.get_board_label(tag)
                    if not trello_label in trello_card_labels_set:
                        trello_card.add_label(trello_label) # Assign label to card
                for label in trello_card_labels_set:
                    if not label.name in taskwarrior_task['tags']:
                        trello_card.remove_label(label) # Remove labels that are not present in tag list
            else:
                # Trello data is newer
                taskwarrior_task['tags'] = trello_card_labels_name_set # Copy tags from Trello labels
                taskwarrior_task_modified = True
            logger.info('Tags of task {} synchronized'.format(taskwarrior_task['id']))
        # Task list name and status <> Trello list name
        if taskwarrior_task.pending and not taskwarrior_task.active and taskwarrior_task['trellolistname'] in [project.trello_doing_list, project.trello_done_list] and taskwarrior_task['modified'] > trello_card.date_last_activity:
            # Task kicked to To Do in Taskwarrior and not synchronized
            trello_card.change_list(self.trello_client.get_list(project.trello_todo_list).id)
            taskwarrior_task['trellolistname'] = list_name = project.trello_todo_list
            taskwarrior_task_modified = True
            logger.info('Task {} kicked to todo list in Trello'.format(taskwarrior_task['id']))
        if taskwarrior_task.active and taskwarrior_task['trellolistname'] != project.trello_doing_list and taskwarrior_task['modified'] > trello_card.date_last_activity:
            # Task kicked to doing in Taskwarrior and not synchronized
            trello_card.change_list(self.trello_client.get_list(project.trello_doing_list).id)
            taskwarrior_task['trellolistname'] = list_name = project.trello_doing_list
            taskwarrior_task_modified = True
            logger.info('Task {} kicked to doing list in Trello'.format(taskwarrior_task['id']))
        if taskwarrior_task.completed and taskwarrior_task['trellolistname'] != project.trello_done_list and taskwarrior_task['modified'] > trello_card.date_last_activity:
            # Task kicked to doing in Taskwarrior and not synchronized
            trello_card.change_list(self.trello_client.get_list(project.trello_done_list).id)
            taskwarrior_task['trellolistname'] = list_name = project.trello_done_list
            taskwarrior_task_modified = True
            logger.info('Task {} kicked to done list in Trello'.format(taskwarrior_task['id']))
        if taskwarrior_task['trellolistname'] != list_name:
            if taskwarrior_task['modified'] > trello_card.date_last_activity:
                # Taskwarrior data is newer
                trello_card.change_list(self.trello_client.get_list(taskwarrior_task['trellolistname']).id)
                logger.info('Task {} kicked to {} list in Trello'.format(taskwarrior_task['id'], taskwarrior_task['trellolistname']))
            else:
                # Trello data is newer
                taskwarrior_task['trellolistname'] = list_name
                if list_name == project.trello_done_list and not taskwarrior_task.completed:
                    taskwarrior_task.save() # Must save before a status change to avoid data loss
                    taskwarrior_task.done()
                    logger.info('Task {} kicked to done list in Taskwarrior'.format(taskwarrior_task['id']))
                elif list_name == project.trello_doing_list:
                    if taskwarrior_task.completed:
                        taskwarrior_task['status'] = 'pending'
                        taskwarrior_task.save()
                        taskwarrior_task.start()
                    elif not taskwarrior_task.active:
                        taskwarrior_task.save()
                        taskwarrior_task.start()
                    else:
                        taskwarrior_task.save()
                    logger.info('Task {} kicked to doing list in Taskwarrior'.format(taskwarrior_task['id']))
                else:
                    if taskwarrior_task.completed:
                        taskwarrior_task['status'] = 'pending'
                        taskwarrior_task.save()
                    elif taskwarrior_task.active:
                        taskwarrior_task.save()
                        taskwarrior_task.stop()
                    else:
                        taskwarrior_task.save()
                    logger.info('Task {} kicked to {} list in Taskwarrior'.format(taskwarrior_task['id'], taskwarrior_task['trellolistname']))
                taskwarrior_task_modified = False # Avoid save again
                logger.info('All changes in Taskwarrior task {} saved'.format(taskwarrior_task['id']))
        # Save Taskwarrior changes (if any)
        if taskwarrior_task_modified:
            taskwarrior_task.save()
            logger.info('All changes in Taskwarrior task {} saved'.format(taskwarrior_task['id']))

    def sync_project(self, project):
        """
        Sync a Taskwarrior project with a Trello board

        :param project: TrelloWarrior project object
        """
        # Initialize clients
        self.taskwarrior_client.project(project)
        self.trello_client.project(project)
        # Get all Taskwarrior deleted tasks and seek for ones that have trelloid (deleted in Taskwarrior)
        logger.info('Syncing project {} step 1: delete Trello cards that already deleted in Taskwarrior'.format(project.name))
        taskwarrior_deleted_tasks = self.taskwarrior_client.get_deleted_tasks()
        for taskwarrior_deleted_task in taskwarrior_deleted_tasks:
            if taskwarrior_deleted_task['trelloid']:
                logger.info('Deleting previously deleted Taskwarrior task with ID {} from Trello'.format(taskwarrior_deleted_task['trelloid']))
                self.trello_client.delete_card(taskwarrior_deleted_task['trelloid'])
                taskwarrior_deleted_task['trelloid'] = None
                taskwarrior_deleted_task.save()
        # Compare and sync Taskwarrior with Trello
        logger.info('Syncing project {} step 2: syncing changes between Taskwarrior and Trello'.format(project.name))
        trello_cards_dict = self.trello_client.get_cards_dict()
        trello_cards_ids = [] # List to store cards IDs to compare later with local trelloid
        for trello_list_name in trello_cards_dict:
            for trello_card in trello_cards_dict[trello_list_name]:
                # Fech all data from card
                trello_card.fetch(False) # Pass False to fetch to avoid download attachments
                trello_cards_ids.append(trello_card.id)
                taskwarrior_task = self.taskwarrior_client.get_task_by_trello_id(trello_card.id)
                if taskwarrior_task is None:
                    # Download new Trello card that not present in Taskwarrior
                    logger.info('Downloading Trello card with ID {} as new task in Taskwarrior'.format(trello_card.id))
                    self.fetch_trello_card(project, trello_list_name, trello_card)
                else:
                    # Sync Taskwarrior task with Trello card
                    self.sync_task_card(project, trello_list_name, trello_card, taskwarrior_task)
        # Compare Trello and Taskwarrior tasks for remove deleted Trello tasks in Taskwarrior
        logger.info('Syncing project {} step 3: delete Takswarrior tasks that already deleted in Trello'.format(project.name))
        taskwarrior_tasks_ids = self.taskwarrior_client.get_tasks_ids_set()
        taskwarrior_tasks_ids.discard(None) # Remove None element if present (new tasks created with Taskwarrior)
        trello_cards_ids = set(trello_cards_ids) # Convert trello_cards_ids list in a set
        for deleted_trello_task_id in taskwarrior_tasks_ids - trello_cards_ids:
            taskwarrior_task_to_delete = self.taskwarrior_client.get_task_by_trello_id(deleted_trello_task_id)
            taskwarrior_task_to_delete['trelloid'] = None
            taskwarrior_task_to_delete.save()
            taskwarrior_task_to_delete.delete()
            logger.info('Deleting previously deleted Trello task with ID {} from Taskwarrior'.format(deleted_trello_task_id))
        # Upload new Taskwarrior tasks that never uploaded before
        logger.info('Syncing project {} step 4: upload new Takswarrior tasks'.format(project.name))
        for taskwarrior_pending_task in self.taskwarrior_client.get_pending_tasks():
            logger.info('Uploading new pending Taskwarrior task with ID {} to Trello'.format(taskwarrior_pending_task['id']))
            if taskwarrior_pending_task.active:
                # Upload new pending active task to doing list
                self.upload_taskwarrior_task(project, taskwarrior_pending_task, self.trello_client.get_list(project.trello_doing_list))
                taskwarrior_pending_task['trellolistname'] = project.trello_doing_list
                taskwarrior_pending_task.save()
            else:
                if taskwarrior_pending_task['trellolistname']:
                    # Upload new pending task to user provided list
                    self.upload_taskwarrior_task(project, taskwarrior_pending_task, self.trello_client.get_list(taskwarrior_pending_task['trellolistname']))
                else:
                    # Upload new pending task to default todo list
                    self.upload_taskwarrior_task(project, taskwarrior_pending_task, self.trello_client.get_list(project.trello_todo_list))
                    taskwarrior_pending_task['trellolistname'] = project.trello_todo_list
                    taskwarrior_pending_task.save()
        for taskwarrior_completed_task in self.taskwarrior_client.get_completed_tasks():
            logger.info('Uploading new completed Taskwarrior task to Trello')
            self.upload_taskwarrior_task(project, taskwarrior_completed_task, self.trello_client.get_list(project.trello_done_list))
            taskwarrior_completed_task['trellolistname'] = project.trello_done_list
            taskwarrior_completed_task.save()
        logger.info('Project {} synchronized'.format(project.name))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2015-2020 Óscar García Amor <ogarcia@connectical.com>
#
# Distributed under terms of the GNU GPLv3 license.

from trellowarrior.commands.auth import auth
from trellowarrior.commands.configedit import config_get, config_set, config_remove
from trellowarrior.commands.configprojectedit import config_project_list, config_project_add, config_project_modify, config_project_show
from trellowarrior.commands.configprojectedit import config_project_enable, config_project_disable, config_project_remove
from trellowarrior.commands.sync import sync
from trellowarrior.commands.version import version
from trellowarrior.config import config

import argparse
import logging
import os

def main():
    # Set loglevel equivalents for argument parser
    log_levels = {
            0: logging.WARNING,
            1: logging.INFO,
            2: logging.DEBUG }

    # Create argument parser to get options via arguments
    parser = argparse.ArgumentParser(description='Bidirectional sync between Trello and Taskwarrior')
    parser.add_argument('-c', '--config', metavar='value', help='custom configuration file path')
    parser.add_argument('-v', '--verbose', action='count', default=0, help='be verbose (add more v to increase verbosity)')
    parser.set_defaults(func=sync) # Perform sync if no command given
    parser.set_defaults(projects=[]) # No forced projects by default
    subparsers = parser.add_subparsers(dest='command')
    subparsers.required = False

    sync_parser = subparsers.add_parser('sync', help='synchronize Trello and Taskwarrior')
    sync_parser.add_argument('projects', nargs='*', help='list of projects to synchronize, if empty will synchronize all enabled projects')
    sync_parser.set_defaults(func=sync)

    auth_parser = subparsers.add_parser('auth', help='setup the authentication against Trello')
    auth_parser.add_argument('--api-key', help='your API Key, can be set from TRELLO_API_KEY environment variable')
    auth_parser.add_argument('--api-key-secret', help='your API Key secret, can be set from TRELLO_API_SECRET environment variable')
    auth_parser.add_argument('--expiration', choices=['1hour', '1day', '30days', 'never'], help='duration of authorization, can be set from TRELLO_EXPIRATION environment variable')
    auth_parser.add_argument('--name', default=os.environ.get('TRELLO_NAME', 'TrelloWarrior'), help='application name, can be set from TRELLO_NAME environment variable, current default: %(default)s')
    auth_parser.set_defaults(func=auth)

    config_parser = subparsers.add_parser('config', help='view or modify TrelloWarrior config')
    config_subparsers = config_parser.add_subparsers(dest='config_subcommand')
    config_subparsers.required = True

    config_get_parser = config_subparsers.add_parser('get', help='get a config value')
    config_get_parser.add_argument('option', help='name of config option, use section.option to get config values outside of default section')
    config_get_parser.set_defaults(func=config_get)

    config_set_parser = config_subparsers.add_parser('set', help='set new or modify a config value')
    config_set_parser.add_argument('option', help='name of config option, use section.option to set config values outside of default section')
    config_set_parser.add_argument('value', help='config value')
    config_set_parser.set_defaults(func=config_set)

    config_remove_parser = config_subparsers.add_parser('remove', help='remove a config value')
    config_remove_parser.add_argument('option', help='name of config option, use section.option to remove config values outside of default section')
    config_remove_parser.set_defaults(func=config_remove)

    config_project_parser = config_subparsers.add_parser('project', help='configure projects')
    config_project_subparsers = config_project_parser.add_subparsers(dest='config_project_subcommand')
    config_project_subparsers.required = True

    config_project_list_parser = config_project_subparsers.add_parser('list', help='list projects')
    config_project_list_parser.set_defaults(func=config_project_list)

    config_project_add_parser = config_project_subparsers.add_parser('add', help='add new project')
    config_project_add_parser.add_argument('name', help='project name')
    config_project_add_parser.add_argument('taskwarrior', help='project name in Taskwarrior')
    config_project_add_parser.add_argument('trello', help='project name in Trello')
    config_project_add_parser.add_argument('--todo', default='To Do', help='to do Trello list name, default: %(default)s')
    config_project_add_parser.add_argument('--doing', default='Doing', help='doing Trello list name, default: %(default)s')
    config_project_add_parser.add_argument('--done', default='Done', help='done Trello list name, default: %(default)s')
    config_project_add_parser.add_argument('--filter', help='Trello lists to filter on sync (separated by commas)')
    config_project_add_parser.add_argument('-o', '--only-my-cards', action='store_true', help='sync only cards assigned to me')
    config_project_add_parser.add_argument('-d', '--disabled', action='store_true', help='add project disabled')
    config_project_add_parser.set_defaults(func=config_project_add)

    config_project_modify_parser = config_project_subparsers.add_parser('modify', help='modify values of existing project')
    config_project_modify_parser.add_argument('name', help='project name')
    config_project_modify_parser.add_argument('--taskwarrior', help='project name in Taskwarrior')
    config_project_modify_parser.add_argument('--trello', help='project name in Trello')
    config_project_modify_parser.add_argument('--todo', help='to do Trello list name')
    config_project_modify_parser.add_argument('--doing', help='doing Trello list name')
    config_project_modify_parser.add_argument('--done', help='done Trello list name')
    config_project_modify_parser.add_argument('--filter', help='Trello lists to filter on sync (separated by commas)')
    config_project_modify_parser.add_argument('--only-my-cards', choices=['yes', 'no'], help='sync only cards assigned to me')
    config_project_modify_parser.set_defaults(func=config_project_modify)

    config_project_show_parser = config_project_subparsers.add_parser('show', help='show project configuration')
    config_project_show_parser.add_argument('name', help='project name')
    config_project_show_parser.set_defaults(func=config_project_show)

    config_project_enable_parser = config_project_subparsers.add_parser('enable', help='enable existing project sync')
    config_project_enable_parser.add_argument('name', help='project name')
    config_project_enable_parser.set_defaults(func=config_project_enable)

    config_project_disable_parser = config_project_subparsers.add_parser('disable', help='disable existing project sync')
    config_project_disable_parser.add_argument('name', help='project name')
    config_project_disable_parser.set_defaults(func=config_project_disable)

    config_project_remove_parser = config_project_subparsers.add_parser('remove', help='remove existing project')
    config_project_remove_parser.add_argument('name', help='project name')
    config_project_remove_parser.set_defaults(func=config_project_remove)

    parser_version = subparsers.add_parser('version', help='show TrelloWarrior version')
    parser_version.set_defaults(func=version)

    args = parser.parse_args()

    # Maximum loglevel is 3 if user sends more vvv we ignore it
    args.verbose = 2 if args.verbose >= 2 else args.verbose

    # Set loglevel via argument or environment (untouched warning by default)
    log_level = log_levels[args.verbose]
    logging.basicConfig(level=log_level)
    logger = logging.getLogger('TrelloWarrior')
    logger.info('Setting loglevel to {}'.format(logging.getLevelName(log_level)))

    # Configure app
    if args.command in ['sync', None]:
        # Need full parse and check configuration for sync (or for no arguments which implies sync)
        config.configure(config_file=args.config, projects=args.projects)
    elif args.command != 'version':
        # Don't need parse condfiguration for rest of commands
        config.configure(config_file=args.config, parse_config_file=False)

    # Run main app
    args.func(args)
